#include <stdio.h>

#include "core/utils/utils.h"
#include "core/debug/evemon.h"
#include "core/debug/exception.h"
#include "core/event/notify.h"
#include "core/scope/base.h"
#include "core/scope/ticket.h"

/* actor class */
class actor : public event::notify::base<>
{
public:
	typedef enum
	{
		EVENT_ACTION=0,
	} event_t;

	actor(void) {}

	virtual void on_notify(int event) override
	{
		_debug("actor::on_notify event=%d", event);
	}

	auto action(void)
	{
		_debug("actor::action");

		return notify(EVENT_ACTION);
	}
};

/* observer class */
class observer : public scope::base
{
public:
	observer(void) {}	

	void init(actor& obj)
	{
		obj.listen(actor::EVENT_ACTION, _SELF(observer, on_action), this);
	}

	void on_action(void)
	{
		_debug("observer::on_action");

		printf("notified using class!\r\n");
	}
};

// main program
int main(int argc, char** argv)
{
	try
	{
		// create actor
		actor obj_actor;

		/*
		 *	example 1
		 *	notify class object
		 */
		printf("example 1\r\n");

		{
			{
				// create observer
				observer obj_observer;

				obj_observer.init(obj_actor);

				// trigger actor action
				printf("%zu observers notified\r\n", obj_actor.action());
			}

			// should fail because scope does not exists anymore
			printf("%zu observers notified\r\n", obj_actor.action());
		}

		printf("\r\n");

		/*
		 *	example 2
		 *	notify without class object
		 */
		printf("example 2\r\n");

		{
			{
				scope::ticket ticket;

				obj_actor.listen(actor::EVENT_ACTION, [](void)
					{
						printf("notify using a lambda!\r\n");
					}, ticket);

				// trigger actor action
				printf("%zu observers notified\r\n", obj_actor.action());
			}

			// should fail because scope does not exists anymore
			printf("%zu observers notified\r\n", obj_actor.action());
		}

		// return successfuly
		return 0;
	}
	catch (exception::base& e)
	{
		printf("error: %s\r\n", e.to_string().c_str());

		_critical("%s", e.to_string().c_str());
	}
	catch (std::exception& e)
	{
		printf("error: %s\r\n", e.what());

		_critical("%s", e.what());
	}
	catch (...)
	{
		printf("unknown error!\r\n");

		_critical("unknown error!");
	}

	// signal error
	return 1;
}